import pydantic

from .base import Base, LowercaseStrEnum


class RadialGridType(LowercaseStrEnum):
    """What sort of radial grid (only one option for now)"""

    KRACK_KOSTER = "krack_koster"


class GridSettings(Base):
    radial_grid_type: RadialGridType = RadialGridType.KRACK_KOSTER
    radial_num_points: pydantic.PositiveInt = 99
    angular_num_points: pydantic.PositiveInt = 590

    weight_cutoff: pydantic.PositiveFloat = 1e-14

    # pruning?
    prune: bool = True
    min_angular_points: pydantic.PositiveInt = 194



from typing import Optional

from .molecule import Molecule
from .base import Base, LowercaseStrEnum
from .settings import Settings
from .status import Status


class StJamesVersion(LowercaseStrEnum):
    """
    Flag for which version of ``stjames`` this calculation was saved in.
    This will become meaningful later, when we have (probably) multiple versions.
    """

    V0 = "09072023"


class Calculation(Base):
    molecules: list[Molecule]

    settings: Settings = Settings()

    status: Status = Status.QUEUED

    name: Optional[str] = None
    elapsed: Optional[float] = None
    logfile: Optional[str] = None

    # not to be changed by end users, diff. versions will have diff. defaults
    json_format: str = StJamesVersion.V0



import pydantic
from typing import Optional, Any

from .modes import Mode
from .methods import Method
from .tasks import Task
from .base import Base
from .corrections import Correction
from .scf_settings import SCFSettings
from .basis_set import BasisSet
from .opt_settings import OptimizationSettings
from .thermochem_settings import ThermochemistrySettings


class Settings(Base):
    method: Method = Method.HARTREE_FOCK
    basis_set: BasisSet = BasisSet(name="STO-3G")
    tasks: list[Task] = [Task.ENERGY, Task.CHARGE, Task.DIPOLE]
    corrections: list[Correction] = []

    mode: Mode = Mode.AUTO

    # scf/opt settings will be set automatically based on mode, but can be overridden manually
    scf_settings: Optional[SCFSettings] = None
    opt_settings: Optional[OptimizationSettings] = None
    thermochem_settings: Optional[ThermochemistrySettings] = ThermochemistrySettings()

    def model_post_init(self, __context: Any) -> None:
        _assign_settings_by_mode(self)

        # composite methods have their own basis sets, so overwrite user stuff
        if self.method == Method.HF3C:
            self.basis_set = BasisSet(name="minix")
        elif self.method == Method.B973C:
            self.basis_set = BasisSet(name="mTZVP")

    @pydantic.field_validator("basis_set", mode="before")
    @classmethod
    def inflate_basis_set(cls, v: Any) -> BasisSet:
        """Turn a string into a ``BasisSettings`` object."""
        if isinstance(v, BasisSet):
            return v
        elif isinstance(v, str):
            return BasisSet(name=v)
        elif v is None:
            return BasisSet(name="STO-3G")
        else:
            raise ValueError(f"invalid value ``{v}`` for ``basis_set``")


def _assign_settings_by_mode(settings: Settings) -> None:
    """Modifies ``scf_settings`` and ``opt_settings`` based on preset ``mode``."""
    mode = settings.mode

    if mode == Mode.AUTO:
        if (
            (Task.OPTIMIZE in settings.tasks)
            or (Task.GRADIENT in settings.tasks)
            or (Task.FREQUENCIES) in settings.tasks
            or (Task.HESSIAN) in settings.tasks
        ):
            mode = Mode.CAREFUL
        else:
            mode = Mode.RAPID

    if settings.scf_settings is None:
        scf_settings = SCFSettings()

        # values based off of the following sources:
        # qchem:
        #   https://manual.q-chem.com/5.2/Ch4.S3.SS2.html
        #   https://manual.q-chem.com/5.2/Ch4.S5.SS2.html
        #
        # gaussian:
        #   https://gaussian.com/integral/
        #   https://gaussian.com/overlay5/
        #
        # orca:
        #   manual 4.2.1, §9.6.1 and §9.7.3
        #
        # psi4:
        #   https://psicode.org/psi4manual/master/autodir_options_c/module__scf.html
        #   https://psicode.org/psi4manual/master/autodoc_glossary_options_c.html
        #
        # terachem:
        #   manual, it's easy to locate everything.
        #
        # the below values are my best attempt at homogenizing various sources.
        # in general, eri_threshold should be 3 OOM lower than scf convergence
        if mode == Mode.RECKLESS:
            scf_settings.energy_threshold = 1e-5
            scf_settings.rms_error_threshold = 1e-7
            scf_settings.max_error_threshold = 1e-4
            scf_settings.rebuild_frequency = 100
            scf_settings.int_settings.eri_threshold = 1e-8
            scf_settings.int_settings.csam_multiplier = 3.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-9
            scf_settings.grid_settings.angular_num_points = 302
            scf_settings.grid_settings.radial_num_points = 75
        elif mode == Mode.RAPID:
            scf_settings.energy_threshold = 1e-6
            scf_settings.rms_error_threshold = 1e-8
            scf_settings.max_error_threshold = 1e-5
            scf_settings.rebuild_frequency = 20
            scf_settings.int_settings.eri_threshold = 1e-9
            scf_settings.int_settings.csam_multiplier = 1.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-10
            scf_settings.grid_settings.angular_num_points = 590
            scf_settings.grid_settings.radial_num_points = 99
        elif mode == Mode.CAREFUL:
            scf_settings.energy_threshold = 1e-7
            scf_settings.rms_error_threshold = 1e-9
            scf_settings.max_error_threshold = 1e-6
            scf_settings.rebuild_frequency = 10
            scf_settings.int_settings.eri_threshold = 1e-10
            scf_settings.int_settings.csam_multiplier = 1.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-10
            scf_settings.grid_settings.angular_num_points = 590
            scf_settings.grid_settings.radial_num_points = 99
        elif mode == Mode.METICULOUS:
            scf_settings.energy_threshold = 1e-8
            scf_settings.rms_error_threshold = 1e-9
            scf_settings.max_error_threshold = 1e-7
            scf_settings.rebuild_frequency = 5
            scf_settings.int_settings.eri_threshold = 1e-12
            scf_settings.int_settings.csam_multiplier = 1.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-12
            scf_settings.grid_settings.angular_num_points = 770
            scf_settings.grid_settings.radial_num_points = 175
        elif mode == Mode.DEBUG:
            scf_settings.energy_threshold = 1e-9
            scf_settings.rms_error_threshold = 1e-10
            scf_settings.max_error_threshold = 1e-9
            scf_settings.rebuild_frequency = 1
            scf_settings.int_settings.eri_threshold = 1e-14
            scf_settings.int_settings.csam_multiplier = (
                1e10  # in other words, disable CSAM
            )
            scf_settings.int_settings.pair_overlap_threshold = 1e-14
            scf_settings.grid_settings.angular_num_points = 974
            scf_settings.grid_settings.radial_num_points = 250
        else:
            raise ValueError(f"Unknown mode ``{mode.value}``!")

        settings.scf_settings = scf_settings

    if settings.opt_settings is None:
        opt_settings = OptimizationSettings()

        # cf. DLFIND manual, and https://www.cup.uni-muenchen.de/ch/compchem/geom/basic.html
        if mode == Mode.RECKLESS:
            opt_settings.energy_threshold = 1e-5
            opt_settings.max_gradient_threshold = 4.5e-3
            opt_settings.rms_gradient_threshold = 3e-3
        elif mode == Mode.RAPID:
            opt_settings.energy_threshold = 1e-6
            opt_settings.max_gradient_threshold = 4.5e-4
            opt_settings.rms_gradient_threshold = 3e-4
        elif mode == Mode.CAREFUL:
            opt_settings.energy_threshold = 1e-6
            opt_settings.max_gradient_threshold = 4.5e-4
            opt_settings.rms_gradient_threshold = 3e-4
        elif mode == Mode.METICULOUS:
            opt_settings.energy_threshold = 1e-7
            opt_settings.max_gradient_threshold = 4.5e-5
            opt_settings.rms_gradient_threshold = 3e-5
        elif mode == Mode.DEBUG:
            opt_settings.energy_threshold = 1e-8
            opt_settings.max_gradient_threshold = 4.5e-6
            opt_settings.rms_gradient_threshold = 3e-6
        else:
            raise ValueError(f"Unknown mode ``{mode.value}``!")

        settings.opt_settings = opt_settings



import pydantic
from enum import Enum
import numpy as np
from typing import Any


class Base(pydantic.BaseModel):
    @pydantic.field_validator("*", mode="before")
    @classmethod
    def coerce_numpy(cls, val: Any):
        if isinstance(val, np.ndarray):
            return val.tolist()
        else:
            return val


class LowercaseStrEnum(str, Enum):
    """Enum where hyphens and case are ignored."""

    @classmethod
    def _missing_(cls, value: str):
        for member in cls:
            if member.lower().replace("-", "") == value.lower().replace("-", ""):
                return member
        return None



import pydantic

from .base import Base


class BasisSet(Base):
    name: str

    # value below which a basis function can be ignored
    # (for improving DFT grid calcs, as per Stratmann/Scuseria/Frisch CPL 1996)
    # this shouldn't really need to be modified...
    cutoff_threshold: pydantic.PositiveFloat = 1e-10



from .base import LowercaseStrEnum


class Method(LowercaseStrEnum):
    #### Hartree–Fock:
    HARTREE_FOCK = "hf"
    HF3C = "hf-3c"

    #### DFT:
    # LDA
    LSDA = "lsda"

    # local GGA
    PBE = "pbe"
    BLYP = "blyp"
    BP86 = "bp86"
    B97D3 = "b97-d3"
    B973C = "b97-3c"

    # hybrid GGA
    PBE0 = "pbe0"
    B3LYP = "b3lyp"
    B3PW91 = "b3pw91"
    PBEH3C = "pbeh-3c"



# ruff: noqa

from .calculation import *
from .molecule import *

from .scf_settings import *
from .int_settings import *
from .opt_settings import *
from .thermochem_settings import *
from .grid_settings import *
from .settings import *

from .methods import *
from .basis_set import *
from .tasks import *
from .corrections import *
from .modes import *
from .status import *



from typing import Optional
import pydantic

from .base import Base, LowercaseStrEnum
from .int_settings import IntSettings
from .grid_settings import GridSettings


class SCFInitMethod(LowercaseStrEnum):
    # (See https://manual.q-chem.com/5.2/Ch4.S4.SS2.html for a nice overview here.)
    SAD = "sad"
    CORE = "core"
    #    GWH = "gwh"
    READ = "read"


class SCFSettings(Base):
    max_iters: int = 100
    init_method: SCFInitMethod = SCFInitMethod.SAD

    int_settings: IntSettings = IntSettings()
    grid_settings: GridSettings = GridSettings()

    #### damping
    do_damping: bool = True
    # when should we stop damping?
    end_damping_error: pydantic.PositiveFloat = 0.5
    # what damping factor should we use?
    damping_factor: float = pydantic.Field(ge=0, le=1, default=0.5)

    #### level shifting
    do_level_shift: bool = True
    # how much? (Eh)
    level_shift_magnitude: pydantic.PositiveFloat = 0.1
    # when should we stop?
    end_level_shift_error: pydantic.PositiveFloat = 0.1

    #### reset incremental fock build
    rebuild_frequency: pydantic.PositiveInt = 20

    #### when are we converged?
    energy_threshold: pydantic.PositiveFloat = 1e-6
    rms_error_threshold: pydantic.PositiveFloat = 1e-8
    max_error_threshold: pydantic.PositiveFloat = 1e-5

    #### DIIS
    do_diis: bool = True
    start_diis_max_error: pydantic.PositiveFloat = 0.5
    start_diis_iter: pydantic.PositiveInt = 3
    diis_subspace_size: pydantic.PositiveInt = 12

    # if ``read`` initialization is selected
    initial_density_matrix_guess: Optional[list[list[float]]] = None



import pydantic
from typing import Self

from .base import Base, LowercaseStrEnum


class ConstraintType(LowercaseStrEnum):
    """Different sorts of constraints."""

    BOND = "bond"
    ANGLE = "angle"
    DIHEDRAL = "dihedral"


class Constraint(Base):
    """Represents a single constraint."""

    constraint_type: ConstraintType
    atoms: list[int]


class OptimizationAlgorithm(LowercaseStrEnum):
    """Which algorithm to use."""

    # default, L-BGFS
    L_BGFS = "l_bgfs"

    # P-RFO, for TS
    P_RFO = "p_rfo"

    # Polak-Ribiere
    CONJUGATE_GRADIENT = "conjugate_gradient"

    # steepest descent
    STEEPEST_DESCENT = "steepest_descent"


class OptimizationSettings(Base):
    max_steps: pydantic.PositiveInt = 100
    algorithm: OptimizationAlgorithm = OptimizationAlgorithm.L_BGFS
    max_stepsize: pydantic.PositiveFloat = 0.05

    # when are we converged?
    max_gradient_threshold: pydantic.PositiveFloat = 4.5e-4
    rms_gradient_threshold: pydantic.PositiveFloat = 3.0e-4
    energy_threshold: pydantic.PositiveFloat = 1e-6

    constraints: list[Constraint] = pydantic.Field(default_factory=list)

    @pydantic.model_validator(mode="after")
    def check_rms_max_grad(self) -> Self:
        rms_grad = self.rms_gradient_threshold
        max_grad = self.max_gradient_threshold
        assert (
            abs(max_grad / rms_grad - 1.50) < 1e-3
        ), "DLFIND hard-codes this relationship between RMS and max gradient!"
        return self



from .base import LowercaseStrEnum


class Task(LowercaseStrEnum):
    ENERGY = "energy"
    GRADIENT = "gradient"
    OPTIMIZE = "optimize"
    CHARGE = "charge"
    SPIN_DENSITY = "spin_density"
    DIPOLE = "dipole"
    HESSIAN = "hessian"
    FREQUENCIES = "frequencies"



import pydantic
from typing import Optional

from .base import Base


class VibrationalMode(Base):
    frequency: float
    reduced_mass: float
    force_constant: float
    displacements: list[list[float]]


class Atom(Base):
    atomic_number: pydantic.NonNegativeInt
    position: list[float]


class Molecule(Base):
    charge: int
    multiplicity: pydantic.PositiveInt
    atoms: list[Atom]

    energy: Optional[float] = None
    scf_iterations: Optional[pydantic.NonNegativeInt] = None
    scf_completed: Optional[bool] = None
    elapsed: Optional[float] = None

    gradient: Optional[list[list[float]]] = None
    density_matrix: Optional[list[list[float]]] = None
    hessian: Optional[list[list[list[list[float]]]]] = None

    atomic_charges: Optional[list[float]] = None
    atomic_spin_densities: Optional[list[float]] = None
    dipole: Optional[list[float]] = None

    vibrational_modes: Optional[list[VibrationalMode]] = None

    zero_point_energy: Optional[float] = None
    thermal_energy_corr: Optional[float] = None
    thermal_enthalpy_corr: Optional[float] = None
    thermal_free_energy_corr: Optional[float] = None

    @property
    def coordinates(self):
        return [a.position for a in self.atoms]

    @property
    def atomic_numbers(self):
        return [a.atomic_number for a in self.atoms]

    @property
    def sum_energy_zpe(self) -> float | None:
        if (self.energy is None) or (self.zero_point_energy is None):
            return None
        else:
            return self.energy + self.zero_point_energy

    @property
    def sum_energy_thermal_corr(self) -> float | None:
        if (self.energy is None) or (self.thermal_energy_corr is None):
            return None
        else:
            return self.energy + self.thermal_energy_corr

    @property
    def sum_energy_enthalpy(self) -> float | None:
        if (self.energy is None) or (self.thermal_enthalpy_corr is None):
            return None
        else:
            return self.energy + self.thermal_enthalpy_corr

    @property
    def sum_energy_free_energy(self) -> float | None:
        if (self.energy is None) or (self.thermal_free_energy_corr is None):
            return None
        else:
            return self.energy + self.thermal_free_energy_corr



import pydantic

from .base import Base, LowercaseStrEnum


class ERIStrategy(LowercaseStrEnum):
    # direct SCF, as per Almlof/Faegri/Korsell
    DIRECT = "direct"

    # Raffinetti-style supermatrix
    SUPERMATRIX = "supermatrix"

    # let the software choose based on the molecule
    AUTO = "auto"


class IntSettings(Base):
    strategy: ERIStrategy = ERIStrategy.AUTO

    # these will get overwritten by ``mode`` anyway, for the most part
    eri_threshold: pydantic.PositiveFloat = 1e-9
    csam_multiplier: pydantic.PositiveInt = pydantic.Field(default=1, ge=1)
    pair_overlap_threshold: pydantic.PositiveFloat = 1e-10
